from setuptools import find_packages, setup

# Read in the version number
exec(open("src/nashpy/version.py", "r").read())

with open("requirements.txt") as f:
    requirements = []
    for library in f.read().splitlines():
        requirements.append(library)


setup(
    name="nashpy",
    version=__version__,
    install_requires=requirements,
    author="Vince Knight, James Campbell",
    author_email=("knightva@cardiff.ac.uk"),
    packages=find_packages("src"),
    package_dir={"": "src"},
    url="",
    license="The MIT License (MIT)",
    description="A library to compute equilibria of 2 player normal form games",
)
